from __future__ import annotations

from pathlib import Path
from weasyprint import HTML
from cstag.to_html import to_html


def to_pdf(cs_tag: str, description: str, path_out: str | Path) -> None:
    """
    Convert a CS tag and its description to a PDF file.

    This function takes a CS (custom string) tag and its description, converts
    it to HTML using the `to_html` function, and then writes it to a PDF file
    using WeasyPrint.

    Args:
        cs_tag (str): The CS tag to be converted.
        description (str): The description associated with the CS tag.
        path_out (str | Path): The path where the output PDF file will be saved.

    Returns:
        None: This function does not return anything.

    Examples:
        >>> import cstag
        >>> cs_tag = "=AC+GGG=T-ACGT*at~gt10cg=GNNN"
        >>> description = "Example"
        >>> path_out = "report.pdf"
        >>> to_pdf(cs_tag, description, path_out)
    """
    cs_tag_html = to_html(cs_tag, description)
    HTML(string=cs_tag_html).write_pdf(path_out)



from __future__ import annotations

from cstag.split import split
from cstag.utils.validator import validate_cs_tag, validate_long_format


def to_sequence(cs_tag: str) -> str:
    """Reconstruct the reference subsequence in the alignment

    Args:
        cs_tag (str): CS tag in the **long** format

    Returns:
        str: The sequence string derived from the CS tag.

    Example:
        >>> import cstag
        >>> cs_tag = "=AC*gt=T-gg=C+tt=A"
        >>> cstag.to_sequence(cs_tag)
        "ACTTCTTA"
    """
    validate_cs_tag(cs_tag)
    validate_long_format(cs_tag)

    cs_tag = cs_tag.replace("cs:Z:", "")
    sequence = []
    for cs in split(cs_tag):
        if cs.startswith("="):
            sequence.append(cs[1:].upper())
        elif cs.startswith("+"):
            sequence.append(cs[1:].upper())
        elif cs.startswith("*"):
            sequence.append(cs[-1].upper())

    return "".join(sequence)



from __future__ import annotations

import re


def split(cs_tag: str, prefix: bool = False) -> list[str]:
    """Split a cs tag
    Args:
        cs_tag (str): a cs tag
        prefix (bool, optional): Whether to add the prefix 'cs:Z:' to the cs tag. Defaults to False
    Return:
        list[str]: splits a cs tag by operators

    Example:
        >>> import cstag
        >>> cs = ":4*ag:3"
        >>> cstag.split(cs)
        [":4", "*ag", ":3"]
    """
    cs_tag = cs_tag.replace("cs:Z:", "")

    pattern = r"(\=[ACGTN]+|:[0-9]+|\*[acgtn][acgtn]|\+[acgtn]+|\-[acgtn]+|\~[acgtn]{2}[0-9]+[acgtn]{2})"
    cs_split = [cs for cs in re.split(pattern, cs_tag) if cs]

    if prefix is True:
        return ["cs:Z:"] + cs_split
    else:
        return cs_split



from __future__ import annotations

import re


def shorten(cs_tag: str, prefix: bool = False) -> str:
    """Convert long format of cs tag into short format
    Args:
        cs_tag (str): cs tag in the **long** format
        prefix (bool, optional): Whether to add the prefix 'cs:Z:' to the cs tag. Defaults to False
    Return:
        str: cs tag in the **short** format
    Example:
        >>> import cstag
        >>> cs = "=ACGT*ag=CGT"
        >>> cstag.shorten(cs, prefix=True)
        cs:Z::4*ag:3
    """
    cstags = re.split(r"([-+*~=])", cs_tag.replace("cs:Z:", ""))[1:]
    cstags = [i + j for i, j in zip(cstags[0::2], cstags[1::2])]

    csshort = []
    for cs in cstags:
        if cs[0] == "=":
            csshort.append(":" + str(len(cs) - 1))
            continue
        csshort.append(cs)
    csshort = "".join(csshort)

    return f"cs:Z:{csshort}" if prefix else csshort



from __future__ import annotations

import re


def validate_cs_tag(cs_tag: str) -> None:
    pattern = re.compile(
        r"^(=[ACGTN]+|:[0-9]+|\*[acgtn][acgtn]|\+[acgtn]+|\-[acgtn]+|\~[acgtn][acgtn][0-9]+[acgtn][acgtn])*$"
    )

    if not pattern.fullmatch(cs_tag.replace("cs:Z:", "")):
        raise ValueError(f"Invalid CS tag: {cs_tag}")


def validate_short_format(cs_tag: str) -> None:
    if re.search(r"=[ACGTN]+", cs_tag):
        raise ValueError("CS tag must be in short format")


def validate_long_format(cs_tag: str) -> None:
    if re.search(r":[0-9]+", cs_tag):
        raise ValueError("CS tag must be in long format")


def validate_threshold(threshold: int) -> None:
    if not isinstance(threshold, int):
        raise ValueError("threshold must be an integer")

    if not 0 <= threshold <= 40:
        raise ValueError("threshold must be within a range between 0 to 40")


def validate_pos(pos: int) -> None:
    if pos < 1:
        raise ValueError(f"pos must be a positive integer, but got {pos}")



from __future__ import annotations

from cstag.shorten import shorten


###########################################################
# Trim soft and hard clips from the CIGAR and sequence
###########################################################


def parse_cigar(cigar: str) -> list[tuple[str, int]]:
    """
    Parse a CIGAR string into a list of tuples containing operation and length.
    """
    parsed_cigar = []
    start_idx = 0
    for idx, operation in enumerate(cigar):
        if operation.isdigit():
            continue
        length = int(cigar[start_idx:idx])
        parsed_cigar.append((operation, length))
        start_idx = idx + 1
    return parsed_cigar


def parse_md(md: str) -> list[tuple[str, int]]:
    """
    Parse an MD tag into a list of tuples containing operation and length.
    """
    parsed_md = []
    idx = 0
    while idx < len(md):
        if md[idx].isdigit():
            start = idx
            while idx < len(md) and md[idx].isdigit():
                idx += 1
            parsed_md.append(("=", int(md[start:idx])))
        elif md[idx] == "^":  # Deletion
            start = idx
            idx += 1
            while idx < len(md) and not md[idx].isdigit():
                idx += 1
            parsed_md.append((md[start:idx], idx - start - 1))
        else:  # Mismatch
            parsed_md.append((md[idx], 1))
            idx += 1
    return parsed_md


def join_cigar(cigar_tuples: list[tuple[str, int]]) -> str:
    return "".join(f"{length}{operation}" for operation, length in cigar_tuples)


def trim_clips(cigar: str, seq: str) -> tuple[str, str]:
    if all(x not in cigar for x in "SH"):
        return cigar, seq
    parsed_cigar = parse_cigar(cigar)
    # trim soft clips of cigar and seq
    if parsed_cigar[0][0] == "S":
        length_softclip = parsed_cigar[0][1]
        seq = seq[length_softclip:]
        parsed_cigar = parsed_cigar[1:]
    if parsed_cigar[-1][0] == "S":
        length_softclip = parsed_cigar[-1][1]
        seq = seq[:-length_softclip]
        parsed_cigar = parsed_cigar[:-1]
    # trim hard clips of cigar
    if parsed_cigar[0][0] == "H":
        parsed_cigar = parsed_cigar[1:]
    if parsed_cigar[-1][0] == "H":
        parsed_cigar = parsed_cigar[:-1]
    cigar = join_cigar(parsed_cigar)
    return cigar, seq


###########################################################
# Generate CS long
###########################################################
def expand_cigar_operations(cigar: str) -> list[str]:
    parsed_cigar = parse_cigar(cigar)
    expanded_list = []
    for op, num in parsed_cigar:
        if op in ["D", "N", "I"]:
            expanded_list.append(op * num)
        else:
            expanded_list.extend([op] * num)
    return expanded_list


def expand_md_operations(md: str) -> list[str]:
    parsed_md = parse_md(md)
    expanded_list = []
    for op, num in parsed_md:
        if op.startswith("^"):
            expanded_list.append(op)
            continue
        expanded_list.extend([op] * num)
    return expanded_list


def remove_consecutive_equals(cs_long: list[str]) -> list[str]:
    if not cs_long:
        return cs_long
    for i, cs in enumerate(cs_long):
        if i == 0:
            prev_op = cs[0]
            continue
        curr_op = cs[0]
        if prev_op == curr_op == "=":
            cs_long[i] = cs[1:]
        else:
            prev_op = cs[0]
    return cs_long


def generate_cs_long(cigar: str, md: str, seq: str) -> str:
    cigar_list = expand_cigar_operations(cigar)
    md_list = expand_md_operations(md)
    idx_cigar, idx_md, idx_seq = 0, 0, 0
    cs_long = []
    while idx_seq < len(seq) and idx_cigar < len(cigar_list) and idx_md < len(md_list):
        if cigar_list[idx_cigar] == "M":
            if md_list[idx_md] == "=":
                cs_long.append(f"={seq[idx_seq]}")
            else:
                cs_long.append(f"*{md_list[idx_md]}{seq[idx_seq]}".lower())
            idx_cigar += 1
            idx_md += 1
            idx_seq += 1
            continue
        if cigar_list[idx_cigar][0] == "D":
            cs_long.append(md_list[idx_md].replace("^", "-").lower())
            idx_cigar += 1
            idx_md += 1
            continue
        if cigar_list[idx_cigar][0] == "I":
            num_insertion = len(cigar_list[idx_cigar])
            seq_insertion = "".join(seq[idx_seq : idx_seq + num_insertion])
            cs_long.append(f"+{seq_insertion}".lower())
            idx_cigar += 1
            idx_seq += num_insertion
            continue
        if cigar_list[idx_cigar][0] == "N":
            cs_long.append(f"~nn{len(cigar_list[idx_cigar])}nn")
            idx_cigar += 1
            continue
    return "".join(remove_consecutive_equals(cs_long))


def add_prefix(cs_tag: str) -> str:
    return f"cs:Z:{cs_tag}"


###########################################################
# main
###########################################################


def call(cigar: str, md: str, seq: str, long: bool = False, prefix: bool = False) -> str:
    """
    Generate a cs tag based on CIGAR, MD, and SEQ information.

    Args:
        cigar (str): CIGAR string representing the alignment.
        md (str): MD tag representing mismatching positions/base.
        seq (str): The sequence of the read.
        long (bool, optional): Whether to return the cs tag in long format. Defaults to False.
        prefix (bool, optional): Whether to add the prefix 'cs:Z:' to the cs tag. Defaults to False

    Returns:
        str: A cs tag representing the alignment and differences.

    Example:
        >>> import cstag
        >>> cigar = "8M2D4M2I3N1M"
        >>> md = "2A5^AG7"
        >>> seq = "ACGTACGTACGTACG"
        >>> cstag.call(cigar, md, seq, long=True)
        '=AC*ag=TACGT-ag=ACGT+ac~nn3nn=G'
    """
    cigar, seq = trim_clips(cigar, seq)
    cs_tag = generate_cs_long(cigar, md, seq)
    if long is False:
        cs_tag = shorten(cs_tag)
    if prefix is True:
        cs_tag = add_prefix(cs_tag)
    return cs_tag



from __future__ import annotations

import re

from cstag.split import split
from cstag.utils.validator import validate_cs_tag, validate_long_format

HTML_HEADER = """<!DOCTYPE html>
    <html>
    <head>
    <style>
    h1 {
    font-family: Consolas, 'Courier New', monospace;
    color: #333;
    padding: 0.1em 0;
    border-top: solid 3px #333;
    border-bottom: solid 3px #333;
    }
    .p_seq {
    font-family: Consolas, 'Courier New', monospace;
    color: #585858;
    word-wrap: break-word;
    letter-spacing: 0.15em;
    }
    .p_legend {
    font-family: Consolas, 'Courier New', monospace;
    color: #585858;
    word-wrap: break-word;
    }

    .Ins {
    color: #333;
    border: 0.1em solid;
    background-color: #ee827c;
    font-weight: bold;
    }
    .Del {
    color: #333;
    border: 0.1em solid;
    background-color: #a0d8ef;
    font-weight: bold;
    }
    .Sub {
    color: #333;
    border: 0.1em solid;
    background-color: #98d98e;
    font-weight: bold;
    }
    .Splice {
    color: #333;
    border: 0.1em solid;
    background-color: #f8e58c;
    font-weight: bold;
    }
    .Unknown {
    color: #333;
    border: 0.1em solid;
    background-color: #c0c6c9;
    font-weight: bold;
    }

    </style>
    </head>
    <body>
"""

HTML_LEGEND = """
<p class = "p_legend">
Labels:
<span class="Ins">Insertion</span>
<span class="Del">Deletion</span>
<span class="Sub">Substitution</span>
<span class="Splice">Splice</span>
<span class="Unknown">Unknown</span>
</p>
<hr>
"""

HTML_FOOTER = """
</body>
</html>
"""


def append_mark_to_n(cs_tag: str) -> str:
    """Process each CS tag by adding specific markers `@` to `N`."""

    def append_mark(cs: str) -> str:
        if cs.startswith("N"):
            return "@" + cs
        elif re.match(r"^[ACGT]", cs):
            return "=" + cs
        return cs

    cs_tag = cs_tag.replace("=N", "N")
    cs_tag_processed = [append_mark(cs) for cs in re.split(r"(N+)", cs_tag) if cs and cs != "="]

    return "".join(cs_tag_processed)


def build_html_parts(cs: str, css_class: str) -> str:
    return f"<span class='{css_class}'>{cs.upper()}</span>"


def process_cs_tag(cs_tag: str) -> str:
    # Format cs_tag
    cs_tag = cs_tag.replace("cs:Z:", "")
    cs_tag_marked = append_mark_to_n(cs_tag)
    cs_tag_split = split(cs_tag_marked)
    # Build html
    html_parts = []
    idx = 0
    while idx < len(cs_tag_split):
        cs = cs_tag_split[idx]
        if cs.startswith("="):
            html_parts.append(cs[1:])
        elif cs.startswith("@"):
            html_parts.append(build_html_parts(cs[1:], "Unknown"))
        elif cs.startswith("*"):
            substitutions = [cs[2]]
            while idx < len(cs_tag_split) - 1 and cs_tag_split[idx + 1].startswith("*"):
                substitutions.append(cs_tag_split[idx + 1][2])
                idx += 1
            html_parts.append(build_html_parts("".join(substitutions), "Sub"))
        elif cs.startswith("+"):
            html_parts.append(build_html_parts(cs[1:], "Ins"))
        elif cs.startswith("-"):
            html_parts.append(build_html_parts(cs[1:], "Del"))
        elif cs.startswith("~"):
            left, right = cs[1:3], cs[-2:]
            splice = "-" * (int(cs[3:-2]) - 4)
            html_parts.append(build_html_parts(f"{left}{splice}{right}", "Splice"))
        idx += 1

    return f"<p class='p_seq'>{''.join(html_parts)}</p>"


def to_html(cs_tag: str, description: str = "") -> str:
    """Output HTML string showing a sequence with mutations colored
    Args:
        cs_tag (str): CS tag in the **long** format
        description (str): (optional) header information in the output string
    Return:
        HTML string
    Example:
        >>> import cstag
        >>> cs_tag = "=AC+GGG=T-ACGT*at~gt10cg=GNNN"
        >>> description = "Example"
        >>> html_string = cstag.to_html(cs_tag, description)
    """
    validate_cs_tag(cs_tag)
    validate_long_format(cs_tag)
    description_str = f"<h1>{description}</h1>" if description else ""
    html_parts = process_cs_tag(cs_tag)
    report = "\n".join(
        [
            HTML_HEADER,
            description_str,
            HTML_LEGEND,
            html_parts,
            HTML_FOOTER,
        ]
    )
    return report



from __future__ import annotations

import re
from cstag.utils.validator import validate_cs_tag, validate_short_format


def lengthen(cs_tag: str, cigar: str, seq: str, prefix: bool = False) -> str:
    """Convert short format of cs tag into long format
    Args:
        cs_tag (str): cs tag in **short** form
        cigar (str): CIGAR string (6th column in SAM file)
        seq (str): segment sequence (10th column in SAM file)
        prefix (bool, optional): Whether to add the prefix 'cs:Z:' to the cs tag. Defaults to False

    Return:
        str: cs tag in **long** form

    Example:
        >>> import cstag
        >>> cs = ":4*ag:3"
        >>> cigar = "8M"
        >>> seq = "ACGTACGT"
        >>> cstag.lengthen(cs, cigar, seq)
        =ACGT*ag=CGT
    """
    validate_cs_tag(cs_tag)
    validate_short_format(cs_tag)

    cs_tag_split = re.split(r"([-+*~:])", cs_tag.replace("cs:Z:", ""))[1:]
    cs_tag_split = [i + j for i, j in zip(cs_tag_split[0::2], cs_tag_split[1::2])]

    softclip = re.sub(r"^([0-9]+)S.*", r"\1", cigar)
    idx = int(softclip) if softclip.isdigit() else 0

    cslong = []
    for cs in cs_tag_split:
        if cs == "":
            continue
        if cs[0] == ":":
            cs = int(cs[1:]) + idx
            cslong.append(":" + seq[idx:cs])
            idx = cs
            continue
        cslong.append(cs)
        if cs[0] == "*":
            idx += 1
        if cs[0] == "+":
            idx += len(cs) - 1
    cslong = "".join(cslong).replace(":", "=")

    return f"cs:Z:{cslong}" if prefix else cslong



"""
.. include:: ../../README.md
"""

from cstag.call import call
from cstag.shorten import shorten
from cstag.lengthen import lengthen
from cstag.consensus import consensus
from cstag.mask import mask
from cstag.split import split
from cstag.revcomp import revcomp
from cstag.to_html import to_html
from cstag.to_pdf import to_pdf
from cstag.to_vcf import to_vcf
from cstag.to_sequence import to_sequence



from __future__ import annotations

import re
from itertools import chain
from collections import deque, Counter

from cstag.utils.validator import validate_cs_tag, validate_long_format


def split_deletion(cs_tag: str) -> list[str]:
    match = re.match(r"-(?P<nucleotides>[acgtn]+)$", cs_tag)
    if match:
        nucleotides = match.group("nucleotides")
        return [f"-{n}" for n in nucleotides]
    return [cs_tag]


def expand_deletion_tags(tags_combined: list[str]) -> list[str]:
    cs_tags_expand_deletion = []
    for tag in tags_combined:
        for tag_splitted in split_deletion(tag):
            cs_tags_expand_deletion.append(tag_splitted)
    return cs_tags_expand_deletion


def split_cs_tags(cs_tags: list[str]) -> list[list[str]]:
    """
    Split and process each CS tag in cs_tags.

    Args:
        cs_tags (list[str]): list of CS tags in the long format.

    Returns:
        list[list[str]]: list of processed CS tags.
    """
    cs_tags_splitted = []
    for cs_tag in cs_tags:
        # Remove the prefix "cs:Z:" if present
        cs_tag = cs_tag.replace("cs:Z:", "")

        # Split the CS tag using special symbols (-, *, ~, =)
        # insertion symbol (+) is ignored because it is not observed in reference sequence
        tags_splitted = re.split(r"([-*~=])", cs_tag)[1:]
        # Combine the symbol with the corresponding sequence
        tags_combined = [symbol + seq for symbol, seq in zip(tags_splitted[0::2], tags_splitted[1::2])]
        tags_combined = expand_deletion_tags(tags_combined)

        # Remove the "=" symbols, as they are not needed for further processing
        cleaned_tags = [tag.replace("=", "") for tag in tags_combined]
        # Further split the tags by the base letters (A, C, G, T)
        further_tags_splitted = [re.split(r"(?=[ACGT])", tag) for tag in cleaned_tags]
        # Remove any empty strings generated by the split
        non_empty_tags = [[elem for elem in tag if elem] for tag in further_tags_splitted]
        # Flatten the list of lists into a single list
        flat_tags = list(chain.from_iterable(non_empty_tags))
        cs_tags_splitted.append(flat_tags)
    return cs_tags_splitted


def normalize_positions(positions: list[int]) -> list[int]:
    """
    Normalize the positions in the given list by shifting them so that the minimum position becomes zero.
    """
    pos_min = min(positions)
    return [pos - pos_min for pos in positions]


def normalize_read_lengths(cs_tags: list[str], positions: list[int]) -> list[list[str]]:
    """
    Normalize the lengths of each read in cs_tags based on their starts positions. If the length is insufficient, fill in with `None`.

    Args:
        cs_tags (list[str]): list of CS tags.
        positions (list[int]): Starting positions of each read.

    Returns:
        list[list[str]]: list of lists representing the reads, now normalized to the same length.
    """
    cs_tags_split = split_cs_tags(cs_tags)
    cs_tags_deque = [deque(cs) for cs in cs_tags_split]
    positions_normalized = normalize_positions(positions)
    cs_maxlen = max(len(cs) + pos for cs, pos in zip(cs_tags_deque, positions_normalized))

    for i, pos in enumerate(positions_normalized):
        if pos > 0:
            cs_tags_deque[i].extendleft([None] * pos)
        if len(cs_tags_deque[i]) < cs_maxlen:
            cs_tags_deque[i].extend([None] * (cs_maxlen - len(cs_tags_deque[i])))
    cs_tags = [list(cs) for cs in cs_tags_deque]
    return cs_tags


def condense_deletions(s: str) -> str:
    # Pattern for detecting continuous nucleotide deletions
    pattern = r"(-[acgtn])+"

    # Function to replace the matched pattern
    def replacement(match) -> str:
        # Remove hyphens and concatenate the nucleotides
        condensed_nucleotides = match.group(0).replace("-", "")
        return f"-{condensed_nucleotides}"

    # Use the regular expression to substitute the pattern with its condensed form
    return re.sub(pattern, replacement, s)


def get_consensus(cs_tags: list[list[str]]) -> str:
    cs_consensus = []
    for cs in zip(*cs_tags):
        # Remove the None that is compensating for the insufficient lead length.
        cs = [c for c in cs if c]
        # Get the most common CS tag(s)
        most_common_tags = Counter(cs).most_common()

        # If there's a unique most common tag, return it
        most_common_tag, _ = most_common_tags[0]
        if len(most_common_tags) == 1 or most_common_tags[0][1] != most_common_tags[1][1]:
            cs_consensus.append(most_common_tag)
            continue
        # If the most common tag is not unique (multimodal), return the first *mutated* mode
        for tag, _ in most_common_tags:
            if not re.search(r"[ACGT]", tag):
                cs_consensus.append(tag)

    cs_consensus = "".join(cs_consensus)
    cs_consensus = condense_deletions(cs_consensus)
    # Append "=" to [ACGTN]
    return re.sub(r"([ACGTN]+)", r"=\1", cs_consensus)


###########################################################
# main
###########################################################


def consensus(cs_tags: list[str], positions: list[int], prefix: bool = False) -> str:
    """generate consensus of CS tags
    Args:
        cs_tags (list): CS tags in the **long** format
        positions (list): 1-based leftmost mapping position (4th column in SAM file)
        prefix (bool, optional): Whether to add the prefix 'cs:Z:' to the CS tag. Defaults to False
    Return:
        str: a consensus of CS tag in the **long** format
    Example:
        >>> import cstag
        >>> cs_tags = ["=ACGT", "=AC*gt=T", "=C*gt=T", "=C*gt=T", "=ACT+ccc=T"]
        >>> positions = [1,1,1,2,1]
        >>> cstag.consensus(cs_tags, positions)
        =AC*gt=T
    """
    if not (len(cs_tags) == len(positions) > 0):
        raise ValueError("Element numbers of each argument must be the same")

    for cs_tag in cs_tags:
        validate_cs_tag(cs_tag)
        validate_long_format(cs_tag)

    cs_tags_normalized_length = normalize_read_lengths(cs_tags, positions)

    cs_consensus = get_consensus(cs_tags_normalized_length)

    return f"cs:Z:{cs_consensus}" if prefix else cs_consensus



from __future__ import annotations

import re

map_revcomp = {
    "A": "T",
    "C": "G",
    "G": "C",
    "T": "A",
    "N": "N",
    "a": "t",
    "c": "g",
    "g": "c",
    "t": "a",
    "n": "n",
}


def _extract_numbers(strings: str) -> list[str]:
    # Using regular expression to find all numbers in the strings: st
    numbers = re.findall(r"\d+", strings)
    return numbers


def revcomp(cs_tag: str, prefix: bool = False) -> str:
    """Converts a cs tag into its reverse complement.
    Args:
        cs_tag (str): a cs tag
        prefix (bool, optional): Whether to add the prefix 'cs:Z:' to the cs tag. Defaults to False

    Return:
        str: reverse complement of a cs tag

    Example:
        >>> import cstag
        >>> cs = "=AAAA*ag=CTG"
        >>> cstag.revcomp(cs)
        =CAG*tc=TTTT
    """
    pattern = r"(\=[ACGTN]+|:[0-9]+|\*[acgtn][acgtn]|\+[acgtn]+|\-[acgtn]+|\~[acgtn]{2}[0-9]+[acgtn]{2})"
    cs_tag_revcomp = []
    cs_tag = cs_tag.replace("cs:Z:", "")
    for cs in re.split(pattern, cs_tag)[::-1]:
        if cs == "":
            continue
        elif cs[0] == ":":
            cs_tag_revcomp.append(cs)
        elif cs[0] == "*":
            cs_tag_revcomp.append(f"*{map_revcomp[cs[1]]}{map_revcomp[cs[2]]}")
        elif cs[0] == "~":
            numbers = _extract_numbers(cs)
            cs_tag_revcomp.append(
                f"~{map_revcomp[cs[-1]]}{map_revcomp[cs[-2]]}{numbers[0]}{map_revcomp[cs[2]]}{map_revcomp[cs[1]]}"
            )
        else:
            op = cs[0]
            cs_revcomp = "".join([map_revcomp[c] for c in cs[1:]])[::-1]
            cs_tag_revcomp.append(f"{op}{cs_revcomp}")
    cs_tag_revcomp = "".join(cs_tag_revcomp)

    if prefix is True:
        return "cs:Z:" + cs_tag_revcomp
    else:
        return cs_tag_revcomp



from __future__ import annotations

import re

from cstag.utils.validator import validate_cs_tag, validate_long_format, validate_threshold


def mask(cs_tag: str, cigar: str, qual: str, threshold: int = 10, prefix: bool = False) -> str:
    """Mask low-quality bases to 'N'
    Args:
        cs_tag (str): cs tag in the **long** format
        cigar (str): cigar strings (6th column in SAM file)
        qual (str): ASCII of Phred-scaled base quaiity+33 (11th column in SAM file)
        threshold (int, optional): Phred Quality Score (defalt = 10). The low-quality bases are defined as 'less than or equal to the threshold'
        prefix (bool, optional): Whether to add the prefix 'cs:Z:' to the cs tag. Defaults to False
    Return:
        str: Masked cs tag
    Example:
        >>> import cstag
        >>> cs_tag = "=ACGT*ac+gg-cc=T"
        >>> cigar = "5M2I2D1M"
        >>> qual = "AA!!!!AA"
        >>> cstag.mask(cs_tag, qual)
        =ACNN*an+ng-cc=T
    """
    validate_cs_tag(cs_tag)
    validate_long_format(cs_tag)
    validate_threshold(threshold)

    mask_symbols = [chr(th + 33) for th in range(threshold + 1)]
    mask_symbols = set(mask_symbols)

    cs = cs_tag.replace("cs:Z:", "")
    list_cs = re.split(r"([-+*~=])", cs)[1:]
    list_cs = [i + j for i, j in zip(list_cs[0::2], list_cs[1::2])]

    if cigar.split("S")[0].isdigit():
        softclip = int(cigar.split("S")[0])
        qual = qual[softclip:]

    cs_masked = []
    idx = 0
    for cs in list_cs:
        cs = list(cs)
        if cs[0] == "*":
            if qual[idx] in mask_symbols:
                cs[-1] = "n"
            idx += 1
        elif cs[0] == "=" or cs[0] == "+":
            for i in range(len(cs) - 1):
                if qual[idx + i] in mask_symbols:
                    cs[i + 1] = "N" if cs[0] == "=" else "n"
            idx += i + 1
        cs_masked.append("".join(cs))
    cs_masked = "".join(cs_masked)

    return f"cs:Z:{cs_masked}" if prefix else cs_masked



from __future__ import annotations

import re
from itertools import chain
from collections import deque, defaultdict, Counter

from dataclasses import dataclass

from cstag.split import split
from cstag.consensus import normalize_read_lengths
from cstag.utils.validator import validate_cs_tag, validate_long_format, validate_pos


@dataclass(frozen=True)
class CsInfo:
    cs_tag: str
    pos_start: int
    pos_end: int
    chrom: str | None = None


@dataclass(frozen=True)
class VcfInfo:
    dp: int | None = None
    rd: int | None = None
    ad: int | None = None
    vaf: float | None = None


@dataclass(frozen=True)
class Vcf:
    chrom: str | None = None
    pos: int | None = None
    ref: str | None = None
    alt: str | None = None
    info: VcfInfo = VcfInfo()


def remove_spaces_around_newlines(text: str) -> str:
    return re.sub(r"\s*\n\s*", "\n", text)


###########################################################
# Get variant annotations
###########################################################


def find_ref_for_insertion(cs_tag_split: list[str], idx: int) -> str | None:
    idx_ref = idx - 1
    while idx_ref >= 0:
        cs = cs_tag_split[idx_ref]
        if cs[0] in ["=", "-"]:
            return cs[-1].upper()
        if cs.startswith("*"):
            return cs[1].upper()
        idx_ref -= 1
    return None


def find_ref_for_deletion(cs_tag_split: list[str], idx: int) -> str:
    ref = deque([cs_tag_split[idx][1:].upper()])
    idx_ref = idx - 1
    while idx_ref >= 0:
        cs = cs_tag_split[idx_ref]
        if cs.startswith("="):
            ref.appendleft(cs[-1].upper())
            break
        if cs.startswith("*"):
            ref.appendleft(cs[1].upper())
            break
        idx_ref -= 1
    return "".join(ref)


def get_variant_annotations(cs_tag_split: list[str], position: int) -> list[Vcf]:
    variant_annotations = []
    pos = position
    for idx, cs in enumerate(cs_tag_split):
        if cs.startswith("="):
            pos += len(cs) - 1
        elif cs.startswith("*"):
            ref, alt = cs[1].upper(), cs[2].upper()
            variant_annotations.append(Vcf(pos=pos, ref=ref, alt=alt))
            pos += 1
        elif cs.startswith("+"):
            ref = find_ref_for_insertion(cs_tag_split, idx)
            alt = ref + cs[1:].upper()
            variant_annotations.append(Vcf(pos=pos - 1, ref=ref, alt=alt))
        elif cs.startswith("-"):
            ref = find_ref_for_deletion(cs_tag_split, idx)
            variant_annotations.append(Vcf(pos=pos - 1, ref=ref, alt=ref[0]))
        elif cs.startswith("~"):
            continue

    return variant_annotations


###########################################################
# Format the CS tags
###########################################################


def get_pos_end(cs_tag: str, pos: int) -> int:
    """Get 1-index end positions"""
    pos_end = pos - 1
    for cs in split(cs_tag):
        if cs[0] in ["=", "-"]:
            pos_end += len(cs) - 1
        if cs[0] == "*":
            pos_end += 1
        else:
            continue
    return pos_end


def format_cs_tags(cs_tags: list[str], chroms: list[str] | list[int], positions: list[int]) -> list[CsInfo]:
    """Format and filter cs_tags, and create a list of CsInfo objects.

    This function takes lists of cs_tags, chromosomes, and positions. It filters
    out any cs_tags containing a tilde ("~") and creates a list of CsInfo objects.

    Args:
        cs_tags (list[str]): List of cs_tags as strings.
        chroms (list[str] | list[int]): List of chromosomes as strings or integers.
        positions (list[int]): List of starting positions as integers.

    Returns:
        list[CsInfo]: A list of CsInfo objects, each containing information about
        a cs_tag, its chromosome, and its start and end positions.
    """

    # Convert all chromosomes to string type
    chroms = [str(chrom) for chrom in chroms]
    # Create a list of CsInfo objects, filtering out any with a splicing ("~") in the cs_tag
    cs_info_list = [
        CsInfo(cs_tag=cs, chrom=chrom, pos_start=pos, pos_end=get_pos_end(cs, pos))
        for cs, chrom, pos in zip(cs_tags, chroms, positions)
        if "~" not in cs
    ]
    return cs_info_list


###########################################################
# Group by chrom and overlapping intervals
###########################################################


def group_by_chrom(cs_tags_formatted: list[tuple]) -> dict[str, tuple]:
    """Group CS tags by chromosomes"""
    cs_tags_grouped = defaultdict(list)
    for cs in cs_tags_formatted:
        cs_tags_grouped[cs.chrom].append(
            CsInfo(cs_tag=cs.cs_tag, pos_start=cs.pos_start, pos_end=cs.pos_end, chrom=cs.chrom)
        )
    return dict(cs_tags_grouped)


def group_by_overlapping_intervals(cs_tags_grouped: CsInfo) -> list[CsInfo]:
    # Sort the list by the starting point
    sorted_data = sorted(cs_tags_grouped, key=lambda x: x.pos_start)
    # Initialize the list of grouped intervals
    grouped_intervals = []
    # Initialize the first group with the first element
    current_group = [sorted_data[0]]
    # Loop through the sorted list starting from the second element
    for i in range(1, len(sorted_data)):
        overlaps = False
        for j in current_group:
            # Check if the intervals overlap
            if sorted_data[i].pos_start <= j.pos_end and sorted_data[i].pos_end >= j.pos_start:
                overlaps = True
                break
        if overlaps:
            # Add the interval to the current group
            current_group.append(sorted_data[i])
        else:
            # Add the current group to the list of grouped intervals
            grouped_intervals.append(current_group)
            # Start a new group
            current_group = [sorted_data[i]]
    # Add the last group to the list of grouped intervals
    grouped_intervals.append(current_group)

    return grouped_intervals


###########################################################
# Add VCF info
###########################################################


def replace_mutation_to_atmark(cs_tags: str) -> str:
    """Replaces mutations with '@'."""
    return "".join(cs if cs in {"A", "C", "G", "T"} else "@" for cs in cs_tags)


def call_reference_depth(variant_annotations, cs_tags_list, positions_list) -> dict[tuple(str, int), int]:
    cs_tags_normalized_length = normalize_read_lengths(cs_tags_list, positions_list)
    cs_replaced = [replace_mutation_to_atmark(cs_tags) for cs_tags in cs_tags_normalized_length]

    reference_depth = defaultdict(int)
    unique_variants = set(variant_annotations)
    for v in unique_variants:
        v_idx = v.pos - min(positions_list)
        for cs in cs_replaced:
            if v.ref == cs[v_idx : v_idx + len(v.ref)]:
                reference_depth[(v.ref, v.pos)] += 1

    return dict(reference_depth)


def add_vcf_fields(
    variant_annotations: list[Vcf], chrom: str, reference_depth: dict[tuple(str, int), int]
) -> list[Vcf]:
    """Add Chrom and VCF info (AD, RD, DP, and VAF) to immutable Vcf dataclass"""
    variant_counter = Counter((v.pos, v.ref, v.alt) for v in variant_annotations)

    updated_annotations = []
    for v in set(variant_annotations):
        ad = variant_counter[(v.pos, v.ref, v.alt)]
        rd = reference_depth.get((v.ref, v.pos), 0)
        dp = rd + ad
        vaf = round(ad / dp, 3) if dp else 0

        # Creating a new VcfInfo object
        updated_info = VcfInfo(dp=dp, rd=rd, ad=ad, vaf=vaf)

        # Creating a new Vcf object
        updated_variant = Vcf(chrom=chrom, pos=v.pos, ref=v.ref, alt=v.alt, info=updated_info)

        updated_annotations.append(updated_variant)

    return updated_annotations


###########################################################
# Process CS tag (One)
###########################################################


def process_cs_tag(cs_tag: str, chrom: str | int, pos: int) -> str:
    validate_cs_tag(cs_tag)
    validate_long_format(cs_tag)
    validate_pos(pos)
    chrom = str(chrom)

    cs_tag_split = split(cs_tag)

    # Call POS, REF, ALT
    variants = get_variant_annotations(cs_tag_split, pos)

    # Write VCF
    HEADER = "##fileformat=VCFv4.2\n#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n"
    vcf = remove_spaces_around_newlines(HEADER).strip().split("\n")
    for v in variants:
        vcf.append(f"{chrom}\t{v.pos}\t.\t{v.ref}\t{v.alt}\t.\t.\t.")

    return "\n".join(vcf)


###########################################################
# Process CS tags (Many)
###########################################################


def chrom_sort_key(chrom: str) -> int:
    """Convert a chromosome string to an integer for sorting."""
    return int(chrom.replace("chr", ""))


def process_cs_tags(cs_tags: list[str], chroms: list[str], positions: list[int]) -> str:
    # validate inputs
    _ = [validate_cs_tag(cs_tag) for cs_tag in cs_tags]
    _ = [validate_long_format(cs_tag) for cs_tag in cs_tags]
    _ = [validate_pos(pos) for pos in positions]

    cs_tags_formatted = format_cs_tags(cs_tags, chroms, positions)
    cs_tags_grouped_by_chrom = group_by_chrom(cs_tags_formatted)

    vcf_info = []
    for chrom, cs_tags_grouped in cs_tags_grouped_by_chrom.items():
        for csinfo in group_by_overlapping_intervals(cs_tags_grouped):
            cs_tags_list = [cs.cs_tag for cs in csinfo]
            positions_list = [cs.pos_start for cs in csinfo]
            variant_annotations = [
                get_variant_annotations(split(cs), pos) for cs, pos in zip(cs_tags_list, positions_list)
            ]
            variant_annotations = list(chain.from_iterable(variant_annotations))
            if not variant_annotations:
                continue
            reference_depth = call_reference_depth(variant_annotations, cs_tags_list, positions_list)
            vcf_info += add_vcf_fields(variant_annotations, chrom, reference_depth)

    # Sort by chrom and pos
    variants = sorted(vcf_info, key=lambda x: (chrom_sort_key(x.chrom), x.pos))

    # Write VCF
    HEADER = """##fileformat=VCFv4.2
    ##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
    ##INFO=<ID=RD,Number=1,Type=Integer,Description="Depth of Ref allele">
    ##INFO=<ID=AD,Number=1,Type=Integer,Description="Depth of Alt allele">
    ##INFO=<ID=VAF,Number=1,Type=Float,Description="Variant allele frequency (AD/DP)">
    #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n
    """

    vcf = remove_spaces_around_newlines(HEADER).strip().split("\n")
    for v in variants:
        vcf.append(
            f"{v.chrom}\t{v.pos}\t.\t{v.ref}\t{v.alt}\t.\t.\tDP={v.info.dp};RD={v.info.rd};AD={v.info.ad};VAF={v.info.vaf}"
        )

    return "\n".join(vcf)


###########################################################
# main
###########################################################


def to_vcf(cs_tags: str | list[str], chroms: str | int | list[str] | list[int], positions: int | list[int]) -> str:
    """
    Convert CS tag(s) to VCF (Variant Call Format) string.

    Args:
        cs_tag (str | list[str]): The CS tag representing the sequence alignment.
        chrom (str | list[str]): The chromosome name.
        pos (int | list[int]): The starting position for the sequence.

    Returns:
        str: The VCF-formatted string.
    Example:
        >>> import cstag
        >>> cs_tag = "=AC*gt=T-gg=C+tt=A"
        >>> chrom = "chr1"
        >>> pos = 1
        >>> print(cstag.to_vcf(cstag, chrom, pos))
        ##fileformat=VCFv4.2
        #CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO
        chr1	3	.	G	T	.	.	.
        chr1	4	.	TGG	T	.	.	.
        chr1	5	.	C	CTT	.	.	.
    """
    if isinstance(cs_tags, str):
        return process_cs_tag(cs_tags, chroms, positions)
    elif isinstance(cs_tags, list):
        return process_cs_tags(cs_tags, chroms, positions)
    else:
        raise TypeError(f"cs_tags must be str or list, not {type(cs_tags)}")
